/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef ABSTRACT_MESSAGE_IO_GATEWAY
#define ABSTRACT_MESSAGE_IO_GATEWAY

#include "dataio/PortableDataIO.h"
#include "message/PortableMessage.h"
#include "queue/PortableQueue.h"
#include "refcount/PortableRefCount.h"
#include "netutil/NetworkUtilityFunctions.h"

/**
 *  Abstract base class representing an object that can send/receive 
 *  PortableMessages via a PortableDataIO byte-stream.
 */
class AbstractMessageIOGateway
{
public:
   /** 
    *  Constructor.
    *  @param io The PortableDataIO object to use to read and/or write data
    *            to/from.  This IOGateway object becomes the owner of (io) and will delete it later.
    */
   AbstractMessageIOGateway(PortableDataIO * io);

   /**
    *  Destructor.
    *  Deletes the held PortableDataIO object.
    */
   virtual ~AbstractMessageIOGateway();

   /**
    * Adds the given message reference to our list of outgoing messages to send.  Never blocks.  
    * @param messageRef A reference to the PortableMessage to send out through the gateway.
    * @return B_NO_ERROR on success, B_ERROR iff for some reason the message can't be queued (out of memory?)
    */
   status_t AddOutgoingMessage(const PortableMessageRef & messageRef);

   /**
    * Retrieves the next message from the incoming messages queue.
    * @param setRef On success, this reference will point to the retrieved PortableMessage.
    * @return B_NO_ERROR on success, or B_ERROR iff the queue is empty.
    */
   status_t GetNextIncomingMessage(PortableMessageRef & setRef);

   /**
    * Writes some of our outgoing message bytes to the wire.
    * Not guaranteed to write all outgoing messages (it will try not to block)
    * or some other catastrophic condition has occurred.
    * @return B_NO_ERROR on success, B_ERROR if something has broken or corrupted the connection.
    */
   virtual status_t DoOutput() = 0;

   /**
    * Reads some more incoming message bytes from the wire.  
    * If enough bytes have been read to assemble one or more new PortableMessages,
    * the new PortableMessages can be retrieved after this method returns by
    * calling GetNextIncomingMessage().  
    * Tries not to block, but may (depending on implementation)
    * @return B_NO_ERROR usually, or B_ERROR if the connection has been broken
    * or some other catastrophic condition has occurred.
    */
   virtual status_t DoInput() = 0;

   /**
    * Returns true If this gateway has bytes that are queued up and waiting
    * to be sent across the wire.  Always returns false if the 
    * connection has been closed or hosed.
    */
   virtual bool HasBytesToOutput() const = 0;

   /**
    * Returns true iff this gateway is interested in new data arriving at its
    * socket.  Returns false if the gateway wants to ignore incoming
    * data for some reason.  Default value is true.
    */
   bool GetInputEnabled() const;

   /**
    * Call this to enable/disable reading of new incoming data from the socket.
    * @param enable If true, the socket will be watched for incoming data.
    *               If false, incoming data will not be read.
    */
   void SetInputEnabled(bool enable);

   /**
    * Returns true iff there are any incoming PortableMessages ready to
    * be retrieved via GetNextIncomingMessage().
    */
   bool HasIncomingMessagesReady() const;

   /**
    * By default, the AbstractMessageIOGateway calls Flush() on its PortableDataIO's
    * output stream whenever the last outgoing message in the outgoing message queue 
    * is sent.  Call SetFlushOnEmpty(false) to inhibit this behavior (e.g. for bandwidth
    * efficiency when low message latency is not a requirement).
    * @param flush If true, auto-flushing will be enabled.  If false, it will be disabled.
    */
   void SetFlushOnEmpty(bool flush);

   /** Accessor for the current state of the FlushOnEmpty flag */
   bool GetFlushOnEmpty() const {return _flushOnEmpty;}

   /**
    * By default, the AbstractMessageIOGateway will use the PortableMessage class's 
    * built-in message and message-ref pool objects to allocate its messages.  This
    * is okay for single-threaded programs, but can lead to race conditions in multi-threaded
    * situations.  In those cases, call SetPools() with your own separate or synchronized
    * pool objects to avoid trouble.
    * The pools may be set to NULL to disable object pooling entirely (i.e. use new/delete instead)
    * @param ioMessagePool An object pool to recycle PortableMessages into, or NULL to disable recycling.
    * @param ioRefPool An object pool to recycle PortableMessagesRefs into, or NULL to disable recycling.
    */
   void SetPools(PortableMessageRef::ItemPool * ioMessagePool, PortableMessageRef::RefPool * ioRefPool);

   /** Returns A reference to our outgoing messages queue.  */
   PortableQueue<PortableMessageRef> & GetOutgoingMessageQueue() {return _outgoingMessages;}

   /** Returns A reference to our incoming messages queue.  */
   PortableQueue<PortableMessageRef> & GetIncomingMessageQueue() {return _incomingMessages;}

   /** Returns A const reference to our outgoing messages queue.  */
   const PortableQueue<PortableMessageRef> & GetOutgoingMessageQueue() const {return _outgoingMessages;}

   /** Returns A const reference to our incoming messages queue.  */
   const PortableQueue<PortableMessageRef> & GetIncomingMessageQueue() const {return _incomingMessages;}

   /** Accessor to our held PortableDataIO */
   PortableDataIO * GetDataIO() {return _io;}

protected:
   /**
    * Convenience method to allocate or reallocate a buffer.  When this method returns
    * successfully, (*bufPtr) will point to a buffer that is at least (desiredSize) bytes long.
    * @param bufPtr points to a pointer to the buffer.  (May point to a NULL
    *               pointer if no buffer hasn't been allocated yet)
    * @param bufSize points to the current size of the buffer.  On return, this value is changed to
    *                reflect the new buffer size.
    * @param desiredSize indicates the minimum new buffer size required.
    * @param copySize indicates the number of bytes to copy out of the
    *                 old buffer and into the new one, if a reallocation is necessary.
    * @return True on success, false on failure (i.e. out of memory).
    */
   bool EnsureBufferSize(uint8 ** bufPtr, size_t * bufSize, size_t desiredSize, size_t copySize);

   /**
    * This can be used if you previously returned a pulse time
    * in GetPulseTime() and have since changed your mind, and don't
    * want to wait until your previously scheduled pulse time to change it.
    * Call this to set the new time that you wish Pulse() to be called at.
    * @param pulseTime the time at which you'd like to be Pulse()'d next.
    */
   void ReschedulePulseTime(uint64 pulseTime);

   /**
    * Same as above, only the new pulse time isn't specified explicitely;
    * rather, it will be determined by calling GetPulseTime().
    */
   void ReschedulePulseTime();       

   /**
    * May be overridden to schedule Pulse times.
    * @param now The current wall-clock time-value, in microseconds, for convenience.
    * @return Return MUSCLE_NEVER_PULSE if you don't wish to schedule a call to Pulse(),
    *         or return the time at which Pulse() should be called.  Return values less
    *         than (now) will cause Pulse() to be called as soon as possible.
    */
   virtual uint64 GetPulseTime(uint64 now);

   /**
    * No-op.  Will be called at the time(s) specified by GetPulseTime().
    * @param now The current wall-clock time-value, in microseconds, for convenience.
    */
   virtual void Pulse(uint64 now);

   /** Returns true iff we are hosed--that is, we've experienced an unrecoverable error. */
   bool IsHosed() const {return _hosed;}

   /** Call this method to flag this gateway as hosed--that is, to say that an unrecoverable error has occurred. */
   void SetHosed() {_hosed = true;}
  
   /** Accessor for our held message pool */
   PortableMessageRef::ItemPool * GetIOMessagePool() {return _ioMessagePool;}

   /** Accessor for our held RefPool */
   PortableMessageRef::RefPool * GetIORefPool() {return _ioRefPool;}
   
private:
   PortableQueue<PortableMessageRef> _outgoingMessages;
   PortableQueue<PortableMessageRef> _incomingMessages;

   PortableDataIO * _io;

   bool _hosed;  // set true on error
   bool _inputEnabled;
   bool _flushOnEmpty;

   PortableMessageRef::ItemPool * _ioMessagePool;
   PortableMessageRef::RefPool * _ioRefPool;

   uint64 _nextPulseAt;

   friend class ReflectServer;
};

#endif
