/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef ABSTRACT_REFLECT_SESSION_H
#define ABSTRACT_REFLECT_SESSION_H

#include <stdio.h>
#include "support/PortableTypeDefs.h"
#include "dataio/TCPSocketDataIO.h"
#include "queue/PortableQueue.h"
#include "refcount/PortableRefCount.h"
#include "iogateway/AbstractMessageIOGateway.h"

class AbstractReflectSession;
class ReflectServer;

/** Type declaration for a function that returns a new AbstractReflectSession object */
typedef AbstractReflectSession* (*ReflectSessionFactoryFunc)();  

/** Type declaration for a reference-count-pointer to an AbstractReflectSession */
typedef PortableRef<AbstractReflectSession> AbstractReflectSessionRef;

/** This is the abstract base class that defines the server side logic for a single
 *  client-server connection.  This class contains no message routing logic of its own, 
 *  but defines the interface so that subclasses can do so.
 */
class AbstractReflectSession
{
public:
   /** Default Constructor. */
   AbstractReflectSession();

   /** Destructor. */
   virtual ~AbstractReflectSession();

   /**
    * This method is called directly after the constructor.
    * Various methods that weren't callable from the constructor
    * may be called here.
    * Should return B_NO_ERROR if everything is okay; something
    * else if there is a problem and the session should be aborted.
    * Default implementation does nothing and returns B_NO_ERROR.
    */
   virtual status_t AttachedToServer();

   /** Returns the hostname of this session */
   const char * GetHostName() const;

   /** Returns the client-side port of this session */
   int GetClientPort() const;

   /**
    * Returns a ID string to represent this session with.
    * The returned string is guaranteed to be unique across
    * all sessions in the server.
    */
   virtual const char * GetSessionIDString() const;

   /**
    * Returns a reference to a PortableMessage that is shared by all sessions
    * This message can be used for whatever purpose the sessions care to; it
    * is not used by the server itself.
    */
   PortableMessage & GetCentralState() const;

   /**
    * Returns an object that can be used to iterate over all the currently
    * connected sessions.
    */
   const PortableQueue<AbstractReflectSessionRef> & GetSessionsList() const;

   /**
    * Returns a reference to the session with the given session ID, or B_ERROR.
    * @param id The ID string of the session you are looking for.
    * @param returnRef on success, a reference to the found session will be copied into this object.
    * @return B_NO_ERROR on success, or B_ERROR if a session with the given session ID couldn't be found.
    */
   status_t GetSessionByID(const char * id, AbstractReflectSessionRef & returnRef) const;
   
   /**
    * Adds the given AbstractReflectSession to the session list.
    * This method is only needed in very specialized situations (e.g. 
    * if sessions are creating new sessions).
    * If (socket) is less than zero, no TCP connection will be used,
    * and the session will be a pure server-side entity.
    * @param session A reference to the new session to add to the server.
    * @param socket the socket descriptor associated with the new session, or -1.
    * @return B_NO_ERROR if the session was successfully added, or B_ERROR on error.
    */
   status_t AddNewSession(AbstractReflectSessionRef session, int socket);

   /** Marks this session to be terminated ASAP, as if the TCP connection had been broken.  */
   void EndSession();

   /** Tells the whole server process to quit ASAP.  */
   void EndServer();

   /**
    * Causes this session to be terminated (similar to EndSession(), 
    * and the session specified in (newSessionRef) to take its 
    * place using the same socket connection & message IO gateway.  
    * @param newSession the new session object that is to take the place of this one.
    * @return B_NO_ERROR on success, B_ERROR if the new session refused to be attached.
    */
   status_t ReplaceSession(AbstractReflectSessionRef newSession);

   /**
    * Called when a new message is received from our IO gateway.
    * @param msg A reference to the PortableMessage that has been received from our client.
    */
   virtual void MessageReceivedFromGateway(PortableMessageRef msg) = 0;

   /**
    * Called when a message is sent to us by another session (possibly this one).  
    * @param from The session who is sending the PortableMessage to us.
    * @param msg A reference to the message that was sent.
    * @param userData Additional data whose semantics are determined by the sending subclass.
    * (For StorageReflectSessions, this value, if non-NULL, is a pointer to the 
    * DataNode in this Session's node subtree that was matched by the paths in (msg))
    */
   virtual void MessageReceivedFromNeighbor(AbstractReflectSession & from, PortableMessageRef msg, void * userData) = 0;

   /**
    * Called when the TCP connection to our client is broken.
    * If this method returns true, then this session will be removed and
    * deleted. 
    * @return If it returns false, then this session will continue, even
    *         though the client is no longer available.  Default implementation always returns true.
    */
   virtual bool ClientConnectionClosed();
  
   /** 
    * Returns the number of milliseconds that the server has been running.
    */
   uint64 GetServerUptime() const;

   /** Passes through to ReflectServer::PutAcceptPort() */
   status_t PutAcceptPort(ReflectSessionFactoryFunc func, short port);

   /** Passes through to ReflectServer::RemoveAcceptPort() */
   void RemoveAcceptPort(short port);

protected:
   /**
    * Adds a PortableMessageRef to our gateway's outgoing messsage queue.
    * (ref) will be sent back to our client when time permits.
    * @param ref Reference to a PortableMessage to send to our client.
    * @return B_NO_ERROR on success, B_ERROR if out-of-memory.
    */
   status_t AddOutgoingMessage(PortableMessageRef ref);

   /**
    * Gateway factory method.  Should return a new AbstractMessageIOGateway
    * to use to communicate over the given socket.  The returned gateway
    * object is responsible for closing the socket when it is done with it.
    * The default implementation returns a regular PortableMessageIOGateway 
    * with non-blocking TCPSocketDataIO.
    * @param socket the socket descriptor to use in the new IO gateway.
    * @return a new message IO gateway that uses the given socket.
    */
   virtual AbstractMessageIOGateway * CreateGateway(int socket);

   /** Should return a pretty, human readable string identifying this class.  */
   virtual const char * GetTypeName() const = 0;

   /** Returns true if we are attached to the ReflectServer object, false if we are not.  */
   bool IsAttachedToServer() const {return (_owner != NULL);}

   /**
    * Returns the held message IO gateway object, or NULL if there is none.
    * The returned object remains the property of this session.
    */
   AbstractMessageIOGateway * GetGateway() const {return _gateway;}

   /** Returns the socket descriptor our gateway is using, or -1.  */
   int GetSocket() const {return _socket;}

   /**
    * Replaces the held message IO gateway object with a new gateway object.
    * @param gw the new AbstractMessageIOGateway to use.
    *       (gw) becomes the property of this session object, and must have been dynamically allocated.
    * @param newSocketFD the TCP socket descriptor that the new gateway will be using, or -1 if (gw) is NULL.
    * @param deleteOldGateway If true, then the old gateway object (if any) will be deleted.   
    */
   void SetGateway(AbstractMessageIOGateway * gw, int newSocketFD, bool deleteOldGateway = true);

   /** 
    * Convenience method:  Calls MessageReceivedFromNeighbor() on all session
    * objects.  Saves you from having to do your own iteration every time you
    * want to broadcast something.
    * @param msgRef a reference to the PortableMessage you wish to broadcast
    * @param userData any userData value you care to include.  Defaults to NULL.
    * @param includeSelf Whether or not MessageReceivedFromNeighbor() should be called on 'this' session.  Defaults to true.
    */
   void BroadcastToAllNeighbors(PortableMessageRef msgRef, void * userData = NULL, bool includeSelf = true);

private:
   friend class ReflectServer;

   char _idString[20];

   // The following variables are set directly by the ReflectServer.
   ReflectServer * _owner;
   int _socket;  
   int _port;
   PortableString _hostName;
   AbstractMessageIOGateway * _gateway; /* we own this object */
};

#endif
