/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <limits.h>
#include "reflector/ReflectServer.h"
#include "reflector/StorageReflectSession.h"
#include "reflector/StorageReflectUtils.h"
#include "stringtokenizer/StringTokenizer.h"

#define DEFAULT_PATH_PREFIX "*/*"  // when we get a path name without a leading '/', prepend this
#define DEFAULT_MAX_SUBSCRIPTION_MESSAGE_SIZE   50   // no more than 50 items/update message, please

static PortableMessageRef _blankMessageRef(newnothrow PortableMessage, NULL, NULL);

StorageReflectSession::DataNodeRef::ItemPool StorageReflectSession::_nodePool(100, StorageReflectSession::ResetNodeFunc);
StorageReflectSession::DataNodeRef::RefPool StorageReflectSession::_nodeRefPool;

StorageReflectSession::StringMatcherRef::ItemPool StorageReflectSession::_stringMatcherPool;
StorageReflectSession::StringMatcherRef::RefPool StorageReflectSession::_stringMatcherRefPool;

StorageReflectSession::StringMatcherQueueRef::ItemPool StorageReflectSession::_stringMatcherQueuePool(100, StorageReflectSession::ResetMatcherQueueFunc);
StorageReflectSession::StringMatcherQueueRef::RefPool StorageReflectSession::_stringMatcherQueueRefPool;

void StorageReflectSession :: DrainPools()
{
   _nodePool.Drain();
   _nodeRefPool.Drain();
   _stringMatcherPool.Drain();
   _stringMatcherRefPool.Drain();
   _stringMatcherQueuePool.Drain();
   _stringMatcherQueueRefPool.Drain();
}

AbstractReflectSession * StorageSessionCreationFunc()
{
   AbstractReflectSession * ret = newnothrow StorageReflectSession;
   if (ret == NULL) WARN_OUT_OF_MEMORY;
   return ret;   
}

StorageReflectSession ::
StorageReflectSession() : 
   _parameters(PR_RESULT_PARAMETERS), 
   _nextSubscriptionMessage(NULL), 
   _nextIndexSubscriptionMessage(NULL), 
   _globalRoot(NULL),
   _subscriptionsEnabled(true), 
   _maxSubscriptionMessageItems(DEFAULT_MAX_SUBSCRIPTION_MESSAGE_SIZE), 
   _indexingPresent(false)
{
   // empty
}

StorageReflectSession :: 
~StorageReflectSession() 
{
   Cleanup();
}

status_t
StorageReflectSession ::
AttachedToServer()
{
   PortableMessage & state = GetCentralState();
   const char * hostname = GetHostName();
   const char * sessionid = GetSessionIDString();

   // First thing, make sure we aren't in the banned list
   const char * nextBan;
   for (int b=0; (state.FindString("banned", b, &nextBan) == B_NO_ERROR); b++)
   {
      if (StringMatcher(nextBan).Match(hostname)) 
      {
         LogTime(MUSCLE_LOG_DEBUG, "Session [%s/%s] matches ban [%s], access denied.\n", hostname, sessionid, nextBan);
         return B_ERROR;      
      }
   }

   // Get the root data node... or add it if it isn't there
   if (state.FindPointer("root", (void **) &_globalRoot) != B_NO_ERROR)
   {
      // oops, there's no global root!  Gotta add it then
      _globalRoot = GetNewDataNode("", _blankMessageRef);
      if (_globalRoot == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
      if (state.AddPointer("root", _globalRoot) != B_NO_ERROR)
      {
         ReleaseDataNode(_globalRoot);
         return B_ERROR;
      }
   }

   // Is there already a node for our hostname?
   DataNodeRef hostDir;
   if (_globalRoot->GetChild(hostname, hostDir) != B_NO_ERROR)
   {
      // nope.... we'll add one then
      DataNode * hostNode = GetNewDataNode(hostname, _blankMessageRef);
      if (hostNode)
      {
         hostDir = DataNodeRef(hostNode, &_nodePool, &_nodeRefPool);
         _globalRoot->PutChild(hostDir, this, this);
      }
      else {WARN_OUT_OF_MEMORY; Cleanup(); return B_ERROR;}
   }

   // Create a new node for our session (we assume no such
   // node already exists, as session id's are supposed to be unique)
   DataNode * hostNode = hostDir.GetItemPointer();
   if (hostNode == NULL) {Cleanup(); return B_ERROR;}
   if (hostNode->HasChild(sessionid)) LogTime(MUSCLE_LOG_WARNING, "WARNING:  Non-unique session id [%s] being overwritten!\n", sessionid);

   DataNode * sessionNode = GetNewDataNode(sessionid, _blankMessageRef);
   if (sessionNode)
   {
      // See if we get any special privileges
      int32 privBits = 0;
      for (int p=0; p<=PR_NUM_PRIVILEGES; p++)
      {
         char temp[32];
         sprintf(temp, "priv%i", p);
         const char * privPattern;
         for (int q=0; (state.FindString(temp, q, &privPattern) == B_NO_ERROR); q++)
         {
            if (StringMatcher(privPattern).Match(hostname))
            {
               if (p == PR_NUM_PRIVILEGES) privBits = ~0;  // all privileges granted!
                                      else privBits |= (1L<<p);
               break;
            } 
         }
      }
      if (privBits != 0L) _parameters.AddInt32(PR_NAME_PRIVILEGE_BITS, privBits);

      _sessionDir = DataNodeRef(sessionNode, &_nodePool, &_nodeRefPool);
      hostNode->PutChild(_sessionDir, this, this);
 
      // do subscription notifications here
      PushSubscriptionMessages();
      return B_NO_ERROR;
   }

   WARN_OUT_OF_MEMORY; 
   Cleanup(); 
   return B_ERROR;
}

// Ensures that we don't leave any droppings behind...
void
StorageReflectSession ::
Cleanup()
{
   if (_globalRoot)
   {
      DataNodeRef hostNodeRef;
      if (_globalRoot->GetChild(GetHostName(), hostNodeRef) == B_NO_ERROR)
      {
         DataNode * hostNode = hostNodeRef.GetItemPointer();
         if (hostNode)
         {
            // make sure our session node is gone
            hostNode->RemoveChild(GetSessionIDString(), this, true);

            // If our host node is now empty, it goes too
            if (hostNode->CountChildren() == 0) _globalRoot->RemoveChild(hostNode->GetNodeName(), this, true);
         }
      
         PushSubscriptionMessages();
      }

      // If the global root is now empty, it goes too
      if (_globalRoot->CountChildren() == 0)
      {
         GetCentralState().RemoveName("root");
         ReleaseDataNode(_globalRoot);
      }
      else 
      {
         // Remove all of our subscription-marks from neighbor's nodes
         _subscriptions.DoTraversal((PathMatchCallback)DoSubscribeRefCallbackFunc, this, *_globalRoot, (void *)(-LONG_MAX));
      }
   }
   GetMessagePool()->ReleaseObject(_nextSubscriptionMessage);
   GetMessagePool()->ReleaseObject(_nextIndexSubscriptionMessage);
}

void 
StorageReflectSession ::
NotifySubscribersThatNodeChanged(DataNode & modifiedNode, bool isBeingRemoved)
{
   PortableHashtableIterator<const char *, uint32> subIter = modifiedNode.GetSubscribers();
   const char * next;
   AbstractReflectSessionRef nRef;
   while(subIter.GetNextKey(next))
   {
      if (GetSessionByID(next, nRef) == B_NO_ERROR) 
      {
         StorageReflectSession * next = (StorageReflectSession *)(nRef.GetItemPointer());
         if ((next)&&((next != this)||(GetReflectToSelf()))) next->NodeChanged(modifiedNode, isBeingRemoved);
      }
   }
}

void 
StorageReflectSession ::
NotifySubscribersThatNodeIndexChanged(DataNode & modifiedNode, char op, uint32 index, const char * key)
{
   PortableHashtableIterator<const char *, uint32> subIter = modifiedNode.GetSubscribers();
   const char * next;
   AbstractReflectSessionRef nRef;
   while(subIter.GetNextKey(next)) 
   {
      if (GetSessionByID(next, nRef) == B_NO_ERROR)
      {
         StorageReflectSession * s = (StorageReflectSession *) nRef.GetItemPointer();
         if (s) s->NodeIndexChanged(modifiedNode, op, index, key);
      }
   }
}

void
StorageReflectSession ::
NotifySubscribersOfNewNode(DataNode & newNode)
{
   const PortableQueue<AbstractReflectSessionRef> & sessions = GetSessionsList();
   for (int i=sessions.CountItems()-1; i>=0; i--)
   {
      StorageReflectSession * next = (StorageReflectSession *) (sessions[i].GetItemPointer());
      if (next) next->NodeCreated(newNode);  // always notify; !Self filtering will be done elsewhere
   }
}

void
StorageReflectSession ::
NodeChanged(DataNode & modifiedNode, bool isBeingRemoved)
{
   if (GetSubscriptionsEnabled())
   {
      if (_nextSubscriptionMessage == NULL) _nextSubscriptionMessage = GetMessageFromPool(PR_RESULT_DATAITEMS);
      if (_nextSubscriptionMessage)
      {
         const char * np = modifiedNode.GetNodePath();
         if (isBeingRemoved) _nextSubscriptionMessage->AddString(PR_NAME_REMOVED_DATAITEMS, np);
                        else _nextSubscriptionMessage->AddMessage(np, modifiedNode.GetData());
         if (_nextSubscriptionMessage->CountNames() >= _maxSubscriptionMessageItems) PushSubscriptionMessages(); 
      }
      else WARN_OUT_OF_MEMORY;
   }
}

void
StorageReflectSession ::
NodeIndexChanged(DataNode & modifiedNode, char op, uint32 index, const char * key)
{
   if (GetSubscriptionsEnabled())
   {
      if (_nextIndexSubscriptionMessage == NULL) _nextIndexSubscriptionMessage = GetMessageFromPool(PR_RESULT_INDEXUPDATED);
      if (_nextIndexSubscriptionMessage)
      {
         char temp[100];
         sprintf(temp, "%c%lu:%s", op, index, key);
         _nextIndexSubscriptionMessage->AddString(modifiedNode.GetNodePath(), temp);
      }
      else WARN_OUT_OF_MEMORY;
      // don't push subscription messages here.... it will be done elsewhere
   }
}

void
StorageReflectSession ::
NodeCreated(DataNode & newNode)
{
   newNode.IncrementSubscriptionRefCount(GetSessionIDString(), _subscriptions.GetMatchCount(newNode));
}

status_t
StorageReflectSession ::
SetDataNode(const PortableString & nodePath, PortableMessageRef dataMsgRef, bool overwrite, bool create)
{
   DataNode * node = _sessionDir.GetItemPointer();
   if (node == NULL) return B_ERROR;
 
   if ((nodePath.Length() > 0)&&(nodePath[0] != '/'))
   {
      int lastSlashPos = -1;
      int slashPos = 0;
      DataNodeRef childNodeRef;
      PortableString nextClause;
      while(slashPos >= 0)
      {
         slashPos = nodePath.IndexOf('/', lastSlashPos+1);
         nextClause = nodePath.Substring(lastSlashPos+1, (slashPos >= 0) ? slashPos : nodePath.Length());
         const char * clause = nextClause.Cstr();
         DataNode * allocedNode = NULL;  // default:  we haven't alloced anybody
         if (node->GetChild(clause, childNodeRef) != B_NO_ERROR)
         {
            if (create)
            {
               // no such child node exists!  Allocate and add it
               allocedNode = GetNewDataNode(clause, _blankMessageRef);
               if (allocedNode)
               {
                  childNodeRef = DataNodeRef(allocedNode, &_nodePool, &_nodeRefPool);
                  node->PutChild(childNodeRef, this, (slashPos < 0) ? NULL : this);  // for last node, notify via SetData(), below
               }
               else {WARN_OUT_OF_MEMORY; return B_ERROR;}
            }
            else return B_ERROR;
         }

         node = childNodeRef.GetItemPointer();
         if (slashPos < 0)
         {
            if ((node == NULL)||((overwrite == false)&&(node != allocedNode))) return B_ERROR;
            node->SetData(dataMsgRef, this);
         }
         lastSlashPos = slashPos;
      }
   }
   return B_NO_ERROR;
}

StorageReflectSession :: DataNode *
StorageReflectSession ::
GetDataNode(const PortableString & nodePath) 
{
   DataNode * node = _sessionDir.GetItemPointer();
   if (nodePath.Length() > 0)
   {
      int lastSlashPos = -1;
      int slashPos = 0;
      DataNodeRef childNodeRef;
      PortableString nextClause;

      if (nodePath[0] == '/') 
      {
         node = _globalRoot;
         lastSlashPos=0;
      }
      while(slashPos >= 0)
      {
         slashPos = nodePath.IndexOf('/', lastSlashPos+1);
         nextClause = nodePath.Substring(lastSlashPos+1, (slashPos >= 0) ? slashPos : nodePath.Length());
         const char * clause = nextClause.Cstr();
         if ((node == NULL)||(node->GetChild(clause, childNodeRef) != B_NO_ERROR)) return NULL;  // lookup failure (404)
         node = childNodeRef.GetItemPointer();
         if (slashPos < 0) return node;
         lastSlashPos = slashPos;
      }
   }
   return node;
}

bool
StorageReflectSession ::
HasPrivilege(int priv) const
{
   int32 privBits;
   return (_parameters.FindInt32(PR_NAME_PRIVILEGE_BITS, &privBits) == B_NO_ERROR) ? ((privBits & (1L<<priv)) != 0L) : false;
}

void 
StorageReflectSession :: 
MessageReceivedFromGateway(PortableMessageRef msgRef)
{
   PortableMessage * msgp = msgRef.GetItemPointer();
   if (msgp == NULL) return;

   PortableMessage & msg = *msgp;
   if ((msg.what >= BEGIN_PR_COMMANDS)&&(msg.what <= END_PR_COMMANDS))
   {
      switch(msg.what)
      {
         case PR_COMMAND_NOOP:
	    // do nothing!
         break;

         case PR_COMMAND_BATCH:
         {
            PortableMessageRef subRef;
            for (int i=0; msg.FindMessage(PR_NAME_KEYS, i, subRef) == B_NO_ERROR; i++) MessageReceivedFromGateway(subRef);
         }
         break;

         case PR_COMMAND_KICK:
         {
            if (HasPrivilege(PR_PRIVILEGE_KICK))
            { 
               if (msg.HasName(PR_NAME_KEYS, B_STRING_TYPE)) 
               {
                  WildPathMatcher matcher;
                  matcher.AddPathsFromMessage(PR_NAME_KEYS, msg, DEFAULT_PATH_PREFIX);
                  matcher.DoTraversal((PathMatchCallback)KickClientCallbackFunc, this, *_globalRoot, NULL);
               }
            }
            else BounceMessage(PR_RESULT_ERRORACCESSDENIED, msgRef);
         }
         break;

         case PR_COMMAND_ADDBANS:
         {
            if (HasPrivilege(PR_PRIVILEGE_ADDBANS))
            {
               const char * nextBan;
               for (int b=0; (msg.FindString(PR_NAME_KEYS, b, &nextBan) == B_NO_ERROR); b++) 
               {
                  LogTime(MUSCLE_LOG_INFO, "Session [%s/%s] is banning [%s]\n", GetHostName(), GetSessionIDString(), nextBan);
                  GetCentralState().AddString("banned", nextBan);
               }
            }
            else BounceMessage(PR_RESULT_ERRORACCESSDENIED, msgRef);
         }
         break;

         case PR_COMMAND_REMOVEBANS:
         {
            if (HasPrivilege(PR_PRIVILEGE_REMOVEBANS))
            {
               PortableMessage & state = GetCentralState();
               const char * pattern;
               for (int b=0; (msg.FindString(PR_NAME_KEYS, b, &pattern) == B_NO_ERROR); b++) 
               {
                  int32 pCount;
                  type_code tc;
                  if (state.GetInfo("banned", &tc, &pCount) == B_NO_ERROR)
                  {
                     StringMatcher sm(pattern);
                     for (int p=pCount-1; p>=0; p--)
                     {
                        const char * ban;
                        if ((state.FindString("banned", p, &ban) == B_NO_ERROR)&&(sm.Match(ban)))
                        {
                           PortableString temp = ban;  // avoid dangling pointer problem in debug output
                           if (state.RemoveData("banned", p) == B_NO_ERROR) LogTime(MUSCLE_LOG_INFO, "Session [%s/%s] removed ban [%s]\n", GetHostName(), GetSessionIDString(), temp.Cstr());
                        }
                     }
                  }
               }
            }
            else BounceMessage(PR_RESULT_ERRORACCESSDENIED, msgRef);
         }
         break;

         case PR_COMMAND_SETPARAMETERS:
         {
            bool subscribeQuietly = msg.HasName(PR_NAME_SUBSCRIBE_QUIETLY);
            PortableMessageFieldNameIterator it = msg.GetFieldNameIterator();
            PortableMessage getMsg(PR_COMMAND_GETDATA);
            const char * fn;
            while((fn = it.GetNextFieldName()) != NULL)
            {
               bool copyField = true;
               if (strncmp(fn, "SUBSCRIBE:", 10) == 0)
               {
                  const char * path = &fn[10];
                  PortableString fixPath(path);
                  _subscriptions.AdjustStringPrefix(fixPath, DEFAULT_PATH_PREFIX);
                  if (_subscriptions.ContainsPathString(fixPath) == false)
                  {
                     // This marks any currently existing matching nodes so they know to notify us
                     // It must be done once per subscription path, as it uses per-sub ref-counting
                     WildPathMatcher temp;
                     temp.AddPathString(fixPath);
                     temp.DoTraversal((PathMatchCallback)DoSubscribeRefCallbackFunc, this, *_globalRoot, (void *)1L);

                     _subscriptions.AddPathsFromMatcher(temp);
                     if (subscribeQuietly == false) getMsg.AddString(PR_NAME_KEYS, path);
                  }
               }
               else if (strcmp(fn, PR_NAME_REFLECT_TO_SELF) == 0)
               {
                  SetReflectToSelf(true);
               }
               else if (strcmp(fn, PR_NAME_DISABLE_SUBSCRIPTIONS) == 0)
               {
                  SetSubscriptionsEnabled(false);
               }
               else if (strcmp(fn, PR_NAME_KEYS) == 0)
               {
                  _defaultMessageRoute.Clear();
                  _defaultMessageRoute.AddPathsFromMessage(PR_NAME_KEYS, msg, DEFAULT_PATH_PREFIX);
               }
               else if (strcmp(fn, PR_NAME_SUBSCRIBE_QUIETLY) == 0)
               {
                  // don't add this to the parameter set; it's just an "argument" for the SUBSCRIBE: fields.
                  copyField = false;
               }
               else if (strcmp(fn, PR_NAME_MAX_UPDATE_MESSAGE_ITEMS) == 0)
               {
                  (void) msg.FindInt32(PR_NAME_MAX_UPDATE_MESSAGE_ITEMS, &_maxSubscriptionMessageItems);
               }
               else if (strcmp(fn, PR_NAME_PRIVILEGE_BITS) == 0)
               {
                  // don't add this to the parameter set; clients aren't allowed to change
                  // their privilege bits (maybe someday, for some clients, but not now)
                  copyField = false;
               }

               if (copyField) msg.CopyName(fn, _parameters);
            }
            if (getMsg.HasName(PR_NAME_KEYS)) DoGetData(getMsg);  // return any data that matches the subscription
         }
         break;

         case PR_COMMAND_GETPARAMETERS:
         {
            DataNode * sessionNode = _sessionDir.GetItemPointer();
            if (sessionNode)
            {
               PortableMessage * resultMessage = GetMessageFromPool(_parameters);
               if (resultMessage)
               {
                  // Add hard-coded params 
                  resultMessage->RemoveName(PR_NAME_SESSION_ROOT);
                  resultMessage->AddString(PR_NAME_SESSION_ROOT, sessionNode->GetNodePath());

                  resultMessage->RemoveName(PR_NAME_SERVER_VERSION);
                  resultMessage->AddString(PR_NAME_SERVER_VERSION, MUSCLE_VERSION_STRING);

                  resultMessage->RemoveName(PR_NAME_SERVER_MEM_AVAILABLE);
                  resultMessage->AddInt64(PR_NAME_SERVER_MEM_AVAILABLE, (uint64) GetNumAvailableBytes());

                  resultMessage->RemoveName(PR_NAME_SERVER_MEM_USED);
                  resultMessage->AddInt64(PR_NAME_SERVER_MEM_USED, (uint64) GetNumUsedBytes());

                  resultMessage->RemoveName(PR_NAME_SERVER_MEM_MAX);
                  resultMessage->AddInt64(PR_NAME_SERVER_MEM_MAX, (uint64) GetMaxNumBytes());

                  resultMessage->RemoveName(PR_NAME_SERVER_UPTIME);
                  resultMessage->AddInt64(PR_NAME_SERVER_UPTIME, (uint64) GetServerUptime());

                  MessageReceivedFromNeighbor(*this, PortableMessageRef(resultMessage, GetMessagePool(), GetMessageRefPool()), NULL);
               }
               else WARN_OUT_OF_MEMORY;
            }
         }
         break;

         case PR_COMMAND_REMOVEPARAMETERS:
         {
            StringMatcher matcher;
            const char * nextName;
            PortableQueue<PortableString> toRemove;

            // Compile a list of parameters whose names match any key in the PR_NAME_KEYS field
            for (int i=0; msg.FindString(PR_NAME_KEYS, i, &nextName) == B_NO_ERROR; i++) 
            {
               // Search the parameters message for all parameters that match (nextName)...
               if (matcher.SetPattern(nextName))
               {
                  PortableMessageFieldNameIterator it = _parameters.GetFieldNameIterator();
                  const char * fn;
                  while((fn = it.GetNextFieldName()) != NULL) if (matcher.Match(fn)) (void) toRemove.AddTail(fn);
               }
               else LogTime(MUSCLE_LOG_ERROR, "PR_COMMAND_REMOVEPARAMETERS:  Error, StringMatcher couldn't match on [%s]\n",nextName);
            }

            // Now remove them from our parameters list (and do any side effects that go with that)
            for (int j=toRemove.CountItems()-1; j>=0; j--) 
            {
               const char * paramName = toRemove[j].Cstr();

               if (_parameters.RemoveName(paramName) == B_NO_ERROR)
               {
                  if (strncmp(paramName, "SUBSCRIBE:", 10) == 0) 
                  {
                     PortableString str(&paramName[10]);
                     _subscriptions.AdjustStringPrefix(str, DEFAULT_PATH_PREFIX);
                     if (_subscriptions.RemovePathString(str) == B_NO_ERROR)
                     {
                        // Remove the references from this subscription from all nodes
                        WildPathMatcher temp;
                        temp.AddPathString(str);
                        temp.DoTraversal((PathMatchCallback)DoSubscribeRefCallbackFunc, this, *_globalRoot, (void *)-1L);
                     }
                  }
                  else if (strcmp(paramName, PR_NAME_REFLECT_TO_SELF) == 0)
                  {
                     SetReflectToSelf(false);
                  }
                  else if (strcmp(paramName, PR_NAME_DISABLE_SUBSCRIPTIONS) == 0)
                  {
                     SetSubscriptionsEnabled(true);
                  }
                  else if (strcmp(paramName, PR_NAME_MAX_UPDATE_MESSAGE_ITEMS) == 0)
                  {
                     _maxSubscriptionMessageItems = DEFAULT_MAX_SUBSCRIPTION_MESSAGE_SIZE;  // back to the default
                  }
               }
            }
         }
         break;

         case PR_COMMAND_SETDATA:
         {
            PortableMessageFieldNameIterator it = msg.GetFieldNameIterator(B_MESSAGE_TYPE);
            const char * nextName;
            while((nextName = it.GetNextFieldName()) != NULL)
            {
               PortableMessageRef dataMsgRef;
               if (msg.FindMessage(nextName, dataMsgRef) == B_NO_ERROR) SetDataNode(nextName, dataMsgRef);
            }
         }
         break;

         case PR_COMMAND_INSERTORDEREDDATA:
         {
            // Because INSERTORDEREDDATA operates solely on pre-existing nodes, we can
            // allow wildcards in our node paths.
            DataNode * sessionNode = _sessionDir.GetItemPointer();
            if (sessionNode)
            {
               WildPathMatcher matcher;
               matcher.AddPathsFromMessage(PR_NAME_KEYS, msg, NULL);
               matcher.DoTraversal((PathMatchCallback)InsertOrderedDataCallbackFunc, this, *sessionNode, msgp);
            }
         }
         break;

         case PR_COMMAND_GETDATA:
            DoGetData(msg);
         break;

         case PR_COMMAND_REMOVEDATA:
         {
            DataNode * sessionNode = _sessionDir.GetItemPointer();
            if (sessionNode)
            {
               PortableQueue<DataNodeRef> removeSet;
               WildPathMatcher matcher;
               matcher.AddPathsFromMessage(PR_NAME_KEYS, msg, NULL);
               matcher.DoTraversal((PathMatchCallback)RemoveDataCallbackFunc, this, *sessionNode, &removeSet);
               for (int i=removeSet.CountItems()-1; i>=0; i--)
               {
                  DataNode * next = removeSet[i].GetItemPointer();
                  DataNode * parent = next->GetParent();
                  if ((next)&&(parent)) parent->RemoveChild(next->GetNodeName(), this, true);
               }
            }
         }
         break;

         case PR_RESULT_PARAMETERS:
            // fall-thru
         case PR_RESULT_DATAITEMS:
            LogTime(MUSCLE_LOG_WARNING, "Warning, client at [%s] sent me a PR_RESULT_* code.  Bad client!\n", GetHostName());
         break;

         case PR_COMMAND_JETTISONRESULTS:
         {
            if (msg.HasName(PR_NAME_KEYS, B_STRING_TYPE))
            {
               WildPathMatcher matcher;
               matcher.AddPathsFromMessage(PR_NAME_KEYS, msg, DEFAULT_PATH_PREFIX);
               JettisonOutgoingResults(&matcher);
            }
            else JettisonOutgoingResults(NULL);
         }
         break;

         case PR_COMMAND_PING:
         {
            msg.what = PR_RESULT_PONG;                         // mark it as processed...
            MessageReceivedFromNeighbor(*this, msgRef, NULL);  // and send it right back to our client
         }
         break;

         default:
            BounceMessage(PR_RESULT_ERRORUNIMPLEMENTED, msgRef);
         break;
      }
   }
   else
   {
      // what code not in our reserved range:  must be a client-to-client message
      if (msg.HasName(PR_NAME_KEYS, B_STRING_TYPE)) 
      {
         WildPathMatcher matcher;
         matcher.AddPathsFromMessage(PR_NAME_KEYS, msg, DEFAULT_PATH_PREFIX);
         matcher.DoTraversal((PathMatchCallback)PassMessageCallbackFunc, this, *_globalRoot, &msgRef);
      }
      else if (_parameters.HasName(PR_NAME_KEYS, B_STRING_TYPE)) 
      {
         _defaultMessageRoute.DoTraversal((PathMatchCallback)PassMessageCallbackFunc, this, *_globalRoot, &msgRef);
      }
      else DumbReflectSession::MessageReceivedFromGateway(msgRef);
   }

   PushSubscriptionMessages();
}

void
StorageReflectSession :: 
BounceMessage(uint32 errorCode, PortableMessageRef msgRef)
{
   // Unknown code; bounce it back to our client
   PortableMessage * bounce = GetMessageFromPool(errorCode);
   if (bounce)
   {
      PortableMessageRef rejRef(bounce, GetMessagePool(), GetMessageRefPool());
      bounce->AddMessage(PR_NAME_REJECTED_MESSAGE, msgRef);
      MessageReceivedFromNeighbor(*this, rejRef, NULL);  // send rejection notice to client
   }
}

void
StorageReflectSession :: 
DoGetData(const PortableMessage & msg)
{
   WildPathMatcher matcher;
   matcher.AddPathsFromMessage(PR_NAME_KEYS, msg, DEFAULT_PATH_PREFIX);  

   PortableMessage * messageArray[2] = {NULL, NULL};  // first entry is the DATAITEMS message, second is the INDEXUPDATED message (both demand-allocated)
   matcher.DoTraversal((PathMatchCallback)GetDataCallbackFunc, this, *_globalRoot, messageArray);

   // Send any still-pending "get" results...
   SendGetDataResults(&messageArray[0]);
   SendGetDataResults(&messageArray[1]);
}

void
StorageReflectSession ::
SendGetDataResults(PortableMessage ** replyMessage)
{
   if (*replyMessage)
   {
      PortableMessageRef msgRef(*replyMessage, GetMessagePool(), GetMessageRefPool());
      MessageReceivedFromNeighbor(*this, msgRef, NULL);
      *replyMessage = NULL;  // force re-alloc later if need be
   }
}


void
StorageReflectSession ::
PushSubscriptionMessages()
{
   // Send out any subscription results that were generated...
   const PortableQueue<AbstractReflectSessionRef> & sessionList = GetSessionsList();
   for (int i=sessionList.CountItems()-1; i>=0; i--)
   {
      StorageReflectSession * nextSession = (StorageReflectSession *) sessionList[i].GetItemPointer();
      if (nextSession)
      {
         nextSession->PushSubscriptionMessage(&nextSession->_nextSubscriptionMessage);
         nextSession->PushSubscriptionMessage(&nextSession->_nextIndexSubscriptionMessage);
      }
   }
}

void
StorageReflectSession ::
PushSubscriptionMessage(PortableMessage ** msgField)
{
   PortableMessage * subMsg = *msgField;  // (msgField) points to one of our _next*SubscriptionMessage members
   while(subMsg) 
   {
      *msgField = NULL;  /* In case FromNeighbor() wants to add more suscriptions... */
      MessageReceivedFromNeighbor(*this, PortableMessageRef(subMsg, GetMessagePool(), GetMessageRefPool()), NULL);
      subMsg = *msgField;
   }
}

int
StorageReflectSession ::
PassMessageCallback(DataNode & node, void * userData)
{
   DataNode * n = &node;
   while(n->GetDepth() > 2) n = n->GetParent();  // go up to session level...
   AbstractReflectSessionRef sref;
   if (GetSessionByID(n->GetNodeName(), sref) == B_NO_ERROR) 
   {
      StorageReflectSession * next = (StorageReflectSession *)(sref.GetItemPointer());
      if ((next)&&((next != this)||(GetReflectToSelf()))) next->MessageReceivedFromNeighbor(*this, *((PortableMessageRef *)userData), &node);
   }
   return 2; // This causes the traversal to immediately skip to the next session
}

int
StorageReflectSession ::
KickClientCallback(DataNode & node, void * /*userData*/)
{
   DataNode * n = &node;
   while(n->GetDepth() > 2) n = n->GetParent();  // go up to session level...
   AbstractReflectSessionRef sref;
   if (GetSessionByID(n->GetNodeName(), sref) == B_NO_ERROR) 
   {
      StorageReflectSession * next = (StorageReflectSession *)(sref.GetItemPointer());
      if ((next)&&(next != this)) 
      {
         LogTime(MUSCLE_LOG_INFO, "Session [%s/%s] is kicking session [%s/%s] off the server.\n", GetHostName(), GetSessionIDString(), next->GetHostName(), next->GetSessionIDString());
         next->EndSession();  // die!!
      }
   }
   return 2; // This causes the traversal to immediately skip to the next session
}

int
StorageReflectSession ::
DoSubscribeRefCallback(DataNode & node, void * userData)
{
   node.IncrementSubscriptionRefCount(GetSessionIDString(), (long) userData);
   return node.GetDepth();  // continue traversal as usual
}

int
StorageReflectSession ::
GetDataCallback(DataNode & node, void * userData)
{
   PortableMessage ** messageArray = (PortableMessage **) userData;

   bool inMyOwnSubtree = false;  // default:  actual value will only be calculated if it makes a difference
   bool reflectToSelf = GetReflectToSelf();
   if (reflectToSelf == false)
   {
      // Make sure (node) isn't part of our own tree!  If it is, move immediately to the next session
      const DataNode * n = &node;
      while(n->GetDepth() > 2) n = n->GetParent();
      AbstractReflectSessionRef ref;
      inMyOwnSubtree = ((GetSessionByID(n->GetNodeName(), ref) == B_NO_ERROR)&&(ref.GetItemPointer() == this));
      if ((_indexingPresent == false)&&(inMyOwnSubtree)) return 2;  // skip to next session node
   }
 
   // Don't send our own data to our own client; he already knows what we have, because he uploaded it!
   if ((inMyOwnSubtree == false)||(reflectToSelf))
   {
      PortableMessage * resultMsg = messageArray[0];
      if (resultMsg == NULL) messageArray[0] = resultMsg = GetMessageFromPool(PR_RESULT_DATAITEMS);
      if (resultMsg)
      {
         resultMsg->AddMessage(node.GetNodePath(), node.GetData());
         if (resultMsg->CountNames() >= _maxSubscriptionMessageItems) SendGetDataResults(&messageArray[0]);
      }
      else 
      {      
         WARN_OUT_OF_MEMORY;
         return 0;  // abort!
      }
   }

   // But indices we need to send to ourself no matter what, as they are generated on the server side.
   const PortableQueue<const char *> & index = node.GetIndex();
   uint32 indexLen = index.CountItems();
   if (indexLen > 0)
   {
      PortableMessage * indexUpdateMsg = messageArray[1];
      if (indexUpdateMsg == NULL) messageArray[1] = indexUpdateMsg = GetMessageFromPool(PR_RESULT_INDEXUPDATED);
      if (indexUpdateMsg)
      {
         char clearStr[] = {INDEX_OP_CLEARED, '\0'};
         indexUpdateMsg->AddString(node.GetNodePath(), clearStr);
         for (uint32 i=0; i<indexLen; i++) 
         {
            char temp[100];
            sprintf(temp, "%c%lu:%s", INDEX_OP_ENTRYINSERTED, i, index[i]);
            indexUpdateMsg->AddString(node.GetNodePath(), temp);
         }
         if (indexUpdateMsg->CountNames() >= _maxSubscriptionMessageItems) SendGetDataResults(&messageArray[1]);
      }
      else 
      {
         WARN_OUT_OF_MEMORY;
         return 0;  // abort!
      }
   }

   return node.GetDepth();  // continue traveral as usual
}

int
StorageReflectSession ::
RemoveDataCallback(DataNode & node, void * userData)
{
   if (node.GetDepth() > 2)  // ensure that we never remove host nodes or session nodes this way
   {
      DataNodeRef nodeRef;
      if (node.GetParent()->GetChild(node.GetNodeName(), nodeRef) == B_NO_ERROR) 
      {
         (void) ((PortableQueue<DataNodeRef> *)userData)->AddTail(nodeRef);
         return node.GetDepth()-1;  // no sense in recursing down a node that we're going to delete anyway
      }
   }
   return node.GetDepth();
}

int
StorageReflectSession ::
InsertOrderedDataCallback(DataNode & node, void * userData)
{
   PortableMessage * insertMsg = (PortableMessage *) userData;
   PortableMessageFieldNameIterator iter = insertMsg->GetFieldNameIterator(B_MESSAGE_TYPE);
   const char * next;
   while((next = iter.GetNextFieldName()) != NULL)
   {
      PortableMessageRef nextRef;
      for (int i=0; (insertMsg->FindMessage(next, i, nextRef) == B_NO_ERROR); i++)
      {
         node.InsertOrderedChild(nextRef, next, this, this);
         _indexingPresent = true;  // disable optimization in GetDataCallback()
      }
   }
   return node.GetDepth();
}


StorageReflectSession :: DataNode ::
DataNode() : _children(CStringHashFunc), _orderedCounter(0L), _subscribers(CStringHashFunc)
{
   _children.SetCompareFunction(CStringCompareFunc);
   _subscribers.SetCompareFunction(CStringCompareFunc);
}

void 
StorageReflectSession :: DataNode ::
Init(const char * name, PortableMessageRef initData)
{
   _nodeName = name;
   _parent   = NULL;
   _depth    = 0;
   _data     = initData;
}

void
StorageReflectSession :: DataNode ::
Reset()
{
   _children.Clear();
   _orderedIndex.MakeEmpty();
   _subscribers.Clear();
   _parent = NULL;
   _depth = 0;
   _data.Reset();
}

void
StorageReflectSession :: DataNode :: 
IncrementSubscriptionRefCount(const char * sessionID, long delta)
{
   if (delta > 0)
   {
      uint32 res = 0;
      (void) _subscribers.Get(sessionID, res);
      (void) _subscribers.Put(sessionID, res+delta);  // I'm not sure how to cleanly handle out-of-mem here??  --jaf
   }
   else if (delta < 0)
   {
      uint32 res = 0;
      if (_subscribers.Get(sessionID, res))
      {
         uint32 decBy = (uint32) -delta;
         if (decBy >= res) _subscribers.Remove(sessionID);
                      else (void) _subscribers.Put(sessionID, res - decBy);  // out-of-mem shouldn't be possible
      }
   }
}

PortableHashtableIterator<const char *, uint32> 
StorageReflectSession :: DataNode ::
GetSubscribers() const
{
   return _subscribers.GetIterator();
}

void
StorageReflectSession :: DataNode ::
InsertOrderedChild(PortableMessageRef data, const char * optInsertBefore, StorageReflectSession * notifyWithOnSetParent, StorageReflectSession * optNotifyChangedData)
{
   // Find a unique ID string for our new kid
   char temp[50];
   while(true)
   {
      sprintf(temp, "I%lu", _orderedCounter++);
      if (HasChild(temp) == false) break;
   }

   DataNode * newNode = notifyWithOnSetParent->GetNewDataNode(temp, data);
   if (newNode == NULL)
   {
      WARN_OUT_OF_MEMORY; 
      return;
   }

   uint32 insertIndex = _orderedIndex.CountItems();  // default to end of index
   if ((optInsertBefore)&&(optInsertBefore[0] == 'I'))  // only 'I''s could be in our index!
   {
      for (int i=_orderedIndex.CountItems()-1; i>=0; i--) 
      {
         if (strcmp(optInsertBefore, _orderedIndex[i]) == 0)
         {
            insertIndex = i;
            break;
         }
      }
   }
 
   // Update the index
   _orderedIndex.InsertItemAt(insertIndex, newNode->GetNodeName());

   DataNodeRef dref(newNode, &_nodePool, &_nodeRefPool);
   PutChild(dref, notifyWithOnSetParent, optNotifyChangedData);

   // Notify anyone monitoring this node that the ordered-index has been updated
   notifyWithOnSetParent->NotifySubscribersThatNodeIndexChanged(*this, INDEX_OP_ENTRYINSERTED, insertIndex, newNode->GetNodeName());
}

void
StorageReflectSession :: DataNode ::
PutChild(DataNodeRef & node, StorageReflectSession * optNotifyWithOnSetParent, StorageReflectSession * optNotifyChangedData)
{
   DataNode * child = node.GetItemPointer();
   if (child)
   {
      child->SetParent(this, optNotifyWithOnSetParent);
      (void) _children.Put(child->_nodeName.Cstr(), node);
      if (optNotifyChangedData) optNotifyChangedData->NotifySubscribersThatNodeChanged(*child, false);
   }
}

StorageReflectSession :: DataNode *
StorageReflectSession :: DataNode ::
GetParent() const
{
   return _parent;
}

int
StorageReflectSession :: DataNode ::
GetDepth() const
{
   return _depth;
}

void
StorageReflectSession :: DataNode ::
SetParent(DataNode * parent, StorageReflectSession * optNotifyWith)
{
   if ((_parent)&&(parent)) LogTime(MUSCLE_LOG_WARNING, "Warning, overwriting previous parent of node [%s]\n", GetNodeName());
   _parent = parent;
   if (_parent == NULL) _subscribers.Clear();
   _depth = 0;

   // Calculate node path and node depth
   if (_parent)
   {
      // Calculate the total length that our node path string will be
      uint32 pathLen = 0;
      {
         const DataNode * node = this;
         while(node->_parent) 
         {
            pathLen += 1 + node->_nodeName.Length();  // the 1 is for the slash
            node = node->_parent;
            _depth++;
         }
      }

      char * aBuf = NULL;
      const uint32 staticAllocSize = 200;
      char sBuf[staticAllocSize];  // try to do this without a dynamic allocation...
      if (pathLen >= staticAllocSize)  // but do a dynamic allocation if we have to (should be rare)
      {
         aBuf = newnothrow char[pathLen+1];
         if (aBuf == NULL) 
         { 
            WARN_OUT_OF_MEMORY; 
            _nodePath = "<error>"; 
            return;
         }
      }

      char * writeAt = (aBuf ? aBuf : sBuf) + pathLen;  // points to last char in buffer
      *writeAt = '\0';  // terminate the string first (!)
      const DataNode * node = this;
      while(node->_parent)
      {
         int len = node->_nodeName.Length();
         writeAt -= len;
         memcpy(writeAt, node->_nodeName.Cstr(), len);
         *(--writeAt) = '/';
         node = node->_parent;
      }
      MASSERT(writeAt == (aBuf ? aBuf : sBuf), "GetNodePath has a bug!");
 
      _nodePath = (aBuf ? aBuf : sBuf);
      delete [] aBuf;

      if (optNotifyWith) optNotifyWith->NotifySubscribersOfNewNode(*this);
   }
}

status_t
StorageReflectSession :: DataNode ::   
GetChild(const char * key, DataNodeRef & returnChild) const
{
   return _children.Get(key, returnChild) ? B_NO_ERROR : B_ERROR;
}

const char * 
StorageReflectSession :: DataNode ::
GetNodePath() const
{
   return _nodePath.Cstr();
}


bool
StorageReflectSession :: DataNode ::   
HasChild(const char * key) const
{
   return _children.ContainsKey(key);
}

status_t
StorageReflectSession :: DataNode ::   
RemoveChild(const char * key, StorageReflectSession * optNotifyWith, bool recurse)
{
   DataNodeRef childRef;
   if (_children.Get(key, childRef))
   {
      DataNode * child = childRef.GetItemPointer();
      if (child)
      {
         if (optNotifyWith) optNotifyWith->NotifySubscribersThatNodeChanged(*child, true);

         // Update our ordered-node index & notify everyone about the change
         if (key[0] == 'I')  // if it doesn't start with I, we know it's not part of our ordered-index!
         {
            for (int i=_orderedIndex.CountItems()-1; i>=0; i--)
            {
               if (strcmp(key, _orderedIndex[i]) == 0)
               {
                  _orderedIndex.RemoveItemAt(i);
                  if (optNotifyWith) optNotifyWith->NotifySubscribersThatNodeIndexChanged(*this, INDEX_OP_ENTRYREMOVED, i, key);
                  break;
               }
            }
         }

         if (recurse)
         { 
            while(child->CountChildren() > 0)
            {
               DataNodeRefIterator it = child->GetChildIterator();
               const char * name;
               it.GetNextKey(name);
               child->RemoveChild(name, optNotifyWith, recurse);
            }
         }

         child->SetParent(NULL, optNotifyWith);
      }
      _children.Remove(key, childRef);
      return B_NO_ERROR;
   }
   return B_ERROR;
}

uint32
StorageReflectSession :: DataNode ::   
CountChildren() const
{
   return _children.Size();
}

const char *
StorageReflectSession :: DataNode ::   
GetNodeName() const
{
   return _nodeName.Cstr();
}

void
StorageReflectSession :: DataNode ::   
SetData(PortableMessageRef data, StorageReflectSession * optNotifyWith)
{
   _data = data;
   if (optNotifyWith) optNotifyWith->NotifySubscribersThatNodeChanged(*this, false);
}

PortableMessageRef 
StorageReflectSession :: DataNode ::
GetData()
{
   return _data;
}

StorageReflectSession :: DataNode ::
~DataNode() 
{
   // empty
}

StorageReflectSession :: DataNodeRefIterator
StorageReflectSession :: DataNode ::
GetChildIterator() const
{
   return _children.GetIterator();
}

StorageReflectSession :: WildPathMatcher ::
WildPathMatcher()
{
   // empty
}

StorageReflectSession :: WildPathMatcher ::
~WildPathMatcher()
{
   // empty
}

void 
StorageReflectSession :: WildPathMatcher :: 
Clear()
{
   _parsers.MakeEmpty();
   _pathStrings.MakeEmpty();
}

void
StorageReflectSession :: WildPathMatcher ::
AdjustStringPrefix(PortableString & path, const char * optPrepend) const
{
   if (path.Length() > 0)
   {
           if (path[0] == '/') path = path.Substring(1);
      else if (optPrepend)     path = optPrepend + ("/" + path);
   }
}

status_t 
StorageReflectSession :: WildPathMatcher ::
AddPathString(const PortableString & path)
{
   if (path.Length() > 0) 
   {
      PortableQueue<StringMatcherRef> * newQ = _stringMatcherQueuePool.GetObject();
      if (newQ == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
      StringMatcherQueueRef qRef(newQ, &_stringMatcherQueuePool, &_stringMatcherQueueRefPool);

      PortableString temp;
      int lastSlashPos = -1;
      int slashPos = 0;
      while(slashPos >= 0)
      {
         slashPos = path.IndexOf('/', lastSlashPos+1);
         temp = path.Substring(lastSlashPos+1, (slashPos >= 0) ? slashPos : path.Length());
         StringMatcher * sm = NULL;  // default, same as "*"
         if (strcmp(temp.Cstr(), "*"))
         {
            sm = _stringMatcherPool.GetObject();
            if (sm == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
            sm->SetPattern(temp.Cstr());
         }
         (void) newQ->AddTail(StringMatcherRef(sm, &StorageReflectSession::_stringMatcherPool, &StorageReflectSession::_stringMatcherRefPool));
         lastSlashPos = slashPos;
      }
      (void) _parsers.AddTail(qRef);
      (void) _pathStrings.AddTail(path);
      return B_NO_ERROR;
   }
   else return B_ERROR;
}

status_t 
StorageReflectSession :: WildPathMatcher ::
RemovePathString(const PortableString & path)
{
   int which = _pathStrings.IndexOf(path);
   if (which >= 0) 
   {
      _parsers.RemoveItemAt(which);
      _pathStrings.RemoveItemAt(which);
      return B_NO_ERROR;
   }
   else LogTime(MUSCLE_LOG_ERROR, "ERROR, COULDN'T REMOVE SUBSCRIBE STRING! [%s]\n", path.Cstr());
   return B_ERROR;
}

bool
StorageReflectSession :: WildPathMatcher ::
ContainsPathString(const PortableString & w) const
{
   return (_pathStrings.IndexOf(w) >= 0);
}

status_t 
StorageReflectSession :: WildPathMatcher ::
AddPathsFromMessage(const char * fieldName, const PortableMessage & msg, const char * prependIfNoLeadingSlash)
{
   status_t ret = B_NO_ERROR;
   PortableString str;
   for (int i=0; msg.FindString(fieldName, i, str) == B_NO_ERROR; i++) 
   {
      AdjustStringPrefix(str, prependIfNoLeadingSlash);
      if (AddPathString(str) != B_NO_ERROR) ret = B_ERROR;
   }
   return ret;
}

void 
StorageReflectSession :: WildPathMatcher ::
AddPathsFromMatcher(const WildPathMatcher & matcher)
{
   for (int i=matcher._parsers.CountItems()-1; i>=0; i--)
   {
      (void) _pathStrings.AddTail(*matcher._pathStrings.GetItemPointer(i));
      (void) _parsers.AddTail(*matcher._parsers.GetItemPointer(i));
   }
}

bool
StorageReflectSession :: WildPathMatcher ::
MatchesNode(DataNode & node) const
{
   for (int i=_parsers.CountItems()-1; i>=0; i--) if (PathMatches(node, i)) return true;
   return false;
}

int
StorageReflectSession :: WildPathMatcher ::
GetMatchCount(DataNode & node) const
{
   int matchCount = 0;
   for (int i=_parsers.CountItems()-1; i>=0; i--) if (PathMatches(node, i)) matchCount++;
   return matchCount;
}

bool
StorageReflectSession :: WildPathMatcher ::
PathMatches(DataNode & node, int whichPath) const
{
   PortableQueue<StringMatcherRef> * nextSubscription = _parsers[whichPath].GetItemPointer();
   if (nextSubscription == NULL) return false;
   int pd = nextSubscription->CountItems();
   if (pd == node.GetDepth())  // only paths with the same number of clauses as the node's path can ever match
   {
      DataNode * travNode = &node;
      bool match = true;  // optimistic default
      for (int j=nextSubscription->CountItems()-1; j>=0; j--,travNode=travNode->GetParent())
      {
         StringMatcher * nextMatcher = nextSubscription->GetItemPointer(j)->GetItemPointer();
         if ((nextMatcher)&&(nextMatcher->Match(travNode->GetNodeName()) == false)) 
         {
            match = false;
            break; 
         }
      }
      if (match) return true;
   }
   return false;
}

bool
StorageReflectSession :: WildPathMatcher :: 
MatchesPath(const char * path) const
{
   uint32 numClauses = GetPathDepth(path);
   for (int i=_parsers.CountItems()-1; i>=0; i--)
   {
      const PortableQueue<StringMatcherRef> * nextSubscription = _parsers.GetItemPointer(i)->GetItemPointer();
      if ((nextSubscription)&&(nextSubscription->CountItems() == numClauses))
      {
         bool matched = true;  // default
         StringTokenizer tok(path, (const char *)"/");
         for (uint32 j=0; j<numClauses; j++)
         {
            const char * nextToken = tok.GetNextToken();
            const StringMatcher * nextMatcher = nextSubscription->GetItemPointer(j)->GetItemPointer();
            if ((nextToken == NULL)||((nextMatcher)&&(nextMatcher->Match(nextToken) == false))) 
            {
               matched = false;
               break;
            }
         }
         if (matched) return true;
      }
   }
   return false;
}

void
StorageReflectSession :: WildPathMatcher ::
DoTraversal(PathMatchCallback cb, StorageReflectSession * This, DataNode & node, void * userData)
{
   (void) DoTraversalAux(cb, This, node, userData, node.GetDepth());
}

int 
StorageReflectSession :: WildPathMatcher ::
DoTraversalAux(PathMatchCallback cb, StorageReflectSession * This, DataNode & node, void * userData, int rootDepth)
{
   DataNodeRef nextChildRef;
   DataNodeRefIterator it = node.GetChildIterator();
   int depth = node.GetDepth();
   while(it.GetNextValue(nextChildRef)) 
   {
      DataNode * nextChild = nextChildRef.GetItemPointer();
      if (nextChild)
      {
         const char * nextChildName = nextChild->GetNodeName();
         bool matched  = false;  // set if we have called the callback on this child already
         bool recursed = false;  // set if we have recursed to this child already

         // Try all parsers and see if any of them match at this level
         int numParsers = _parsers.CountItems();
         for (int i=0; i<numParsers; i++)
         {
            PortableQueue<StringMatcherRef> * nextQueue = _parsers[i].GetItemPointer();
            if (nextQueue)
            {
               int numClausesInParser = nextQueue->CountItems();
               if (numClausesInParser > depth-rootDepth)
               {
                  StringMatcher * nextMatcher = nextQueue->GetItemPointer(depth-rootDepth)->GetItemPointer();
                  if ((nextMatcher == NULL)||(nextMatcher->Match(nextChildName)))
                  {
                     // A match!  Now, depending on whether this match is the
                     // last clause in the path or not, we either do the callback or descend.
                     // But we make sure not to do either of these things more than once per node.

                     if (depth == rootDepth+numClausesInParser-1)
                     {
                        if (matched == false)
                        {
                           // when there is more than one string being used to match,
                           // it's possible that two or more strings can "conspire"
                           // to match a node even though any given string doesn't match it.
                           // For example, if we have the match-strings:
                           //    /j*/k*
                           //    /k*/j*
                           // The node /jeremy/jenny would match, even though it isn't
                           // specified by any of the subscription strings.  This is bad.
                           // So for multiple match-strings, we do an additional check 
                           // to make sure there is a WildPathMatcher for this node.
                           if ((numParsers == 1)||(MatchesNode(*nextChild)))
                           {
                              int nextDepth = cb(This, *nextChild, userData);
                              if (nextDepth < nextChild->GetDepth()-1) return nextDepth;
                              matched = true;
                              if (recursed) break;  // done both possibile actions, so be lazy
                           }
                        }
                     }
                     else
                     {
                        if (recursed == false)
                        {
                           // If we match a non-terminal clause in the path, recurse to the child.
                           int nextDepth = DoTraversalAux(cb, This, *nextChild, userData, rootDepth);
                           if (nextDepth < nextChild->GetDepth()-1) return nextDepth;
                           recursed = true;
                           if (matched) break;  // done both possible actions, so be lazy
                        }
                     }
                  }
               }
            }
         }
      }
   }
   return node.GetDepth();
}

StorageReflectSession::DataNode *
StorageReflectSession ::
GetNewDataNode(const char * name, PortableMessageRef initialValue)
{
   DataNode * n = _nodePool.GetObject();
   if (n) n->Init(name, initialValue);
   return n;
}

void
StorageReflectSession ::
ReleaseDataNode(DataNode * node)
{
   _nodePool.ReleaseObject(node);
}

void
StorageReflectSession ::
JettisonOutgoingResults(const WildPathMatcher * matcher)
{
   AbstractMessageIOGateway * gw = GetGateway();
   if (gw)
   {
      PortableQueue<PortableMessageRef> & oq = gw->GetOutgoingMessageQueue();
      for (int i=oq.CountItems()-1; i>=0; i--)  // must do this backwards!
      {
         PortableMessage * msg = oq.GetItemPointer(i)->GetItemPointer();
         if ((msg)&&(msg->what == PR_RESULT_DATAITEMS))
         {
            if (matcher)
            {
               // Remove any PR_NAME_REMOVED_DATAITEMS entries that match... 
               bool hasRemoveItems = false;  // will be set true iff any remove-item entries remain
               int nextr = 0;
               const char * rname;
               while(msg->FindString(PR_NAME_REMOVED_DATAITEMS, nextr, &rname) == B_NO_ERROR)
               {
                  if (matcher->MatchesPath(rname)) msg->RemoveData(PR_NAME_REMOVED_DATAITEMS, nextr);
                  else
                  {
                     hasRemoveItems = true; 
                     nextr++;
                  }
               }
   
               // Build up a list of all update fields to remove (we can't remove them
               // immediately because it would break the iterator)
               PortableQueue<PortableString> toRemove;
               const char * nextFieldName;
               PortableMessageFieldNameIterator iter = msg->GetFieldNameIterator(B_MESSAGE_TYPE);
               bool hasUpdateFields = false;
               while((nextFieldName = iter.GetNextFieldName()) != NULL) 
               {
                  if (matcher->MatchesPath(nextFieldName)) (void) toRemove.AddTail(nextFieldName);
                                                      else hasUpdateFields = true;
               }
   
               if ((hasRemoveItems)||(hasUpdateFields))
               {
                  // Get rid of the matched fields only
                  for (int j=toRemove.CountItems()-1; j>=0; j--) msg->RemoveName(toRemove.GetItemPointer(i)->Cstr());
               }
               else oq.RemoveItemAt(i);  // nothing left, so just dump the whole message
            }
            else oq.RemoveItemAt(i);
         }
      }
   }
}
